import datetime

from itsdangerous import BadSignature
from itsdangerous.url_safe import URLSafeTimedSerializer
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy import select, insert
from quart import g

from .user import User, Token
from foxnouns.exceptions import ForbiddenError, ErrorCode
from foxnouns.settings import SECRET_KEY


async def user_from_ref(session: AsyncSession, user_ref: str):
    """Returns a user from a `user_ref` value. If `user_ref` is `@me`, returns the current user.
    Otherwise, tries to convert the user to a snowflake ID and queries that. Otherwise, returns a user with that username.
    """
    query = select(User)

    if user_ref == "@me":
        if "user" in g:
            if g.token.has_scope("user.read"):
                query = query.where(User.id == g.user.id)
            else:
                raise ForbiddenError(
                    f"Missing scope 'user.read'", type=ErrorCode.MissingScope
                )
        else:
            raise ForbiddenError("Not authenticated")
    else:
        try:
            id = int(user_ref)
            query = query.where(User.id == id)
        except ValueError:
            query = query.where(User.username == user_ref)

    return await session.scalar(query)


serializer = URLSafeTimedSerializer(SECRET_KEY)


def generate_token(token: Token):
    return serializer.dumps(token.id)


async def create_token(session: AsyncSession, user: User, scopes: list[str] = ["*"]):
    expires = datetime.datetime.now() + datetime.timedelta(days=90)
    query = (
        insert(Token)
        .values(user_id=user.id, expires_at=expires, scopes=scopes)
        .returning(Token)
    )
    return await session.scalar(query)


async def validate_token(session: AsyncSession, header: str) -> (Token, User):
    try:
        token_id = serializer.loads(header)
    except BadSignature:
        raise ForbiddenError("Invalid token", type=ErrorCode.InvalidToken)

    row = (
        await session.execute(
            select(Token, User).join(Token.user).where(Token.id == token_id)
        )
    ).first()

    if not row or not row.Token:
        raise ForbiddenError("Invalid token", type=ErrorCode.InvalidToken)

    if row.Token.expires_at < datetime.datetime.now():
        raise ForbiddenError("Token has expired", type=ErrorCode.InvalidToken)

    return (row.Token, row.User)


def is_self(user: User) -> bool:
    return "user" in g and g.user.id == user.id
